"""
The pytoil info command.

Author: Tom Fleet
Created: 25/06/2021
"""

import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.exceptions import RepoNotFoundError
from pytoil.repo import Repo

app = typer.Typer()


@app.command()
def info(
    project: str = typer.Argument(..., help="Name of the project to fetch info for.")
) -> None:
    """
    Get useful info for a project.

    Given a project name (can be local or remote), 'info' will return a summary
    description of the project.

    If the project is on GitHub, info will prefer getting information from the GitHub
    API as this is more detailed.

    If the project is local only, some information is extracted from the operating
    system about the project.

    Examples:

    $ pytoil info my_project
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )

    try:
        _ = repo.exists_remote(api=api)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)

    try:
        with msg.loading(f"Fetching info for {project!r}..."):
            info_dict = repo.info(api=api)

    except RepoNotFoundError:
        msg.warn(
            f"{project!r} not found locally or on GitHub. Was it a typo?",
            spaced=True,
            exits=1,
        )
    else:
        typer.secho(f"\nInfo for {project!r}:\n", fg=typer.colors.CYAN, bold=True)
        for key, val in info_dict.items():
            typer.echo(f"{key}: {val}")



"""
Keeps the Logo out of the way of any logic in root.py.

Author: Tom Fleet
Created: 02/07/2021
"""


import platform
import sys
from typing import Dict

import typer

from pytoil import __version__

LOGO = r"""
██████╗ ██╗   ██╗████████╗ ██████╗ ██╗██╗
██╔══██╗╚██╗ ██╔╝╚══██╔══╝██╔═══██╗██║██║
██████╔╝ ╚████╔╝    ██║   ██║   ██║██║██║
██╔═══╝   ╚██╔╝     ██║   ██║   ██║██║██║
██║        ██║      ██║   ╚██████╔╝██║███████╗
╚═╝        ╚═╝      ╚═╝    ╚═════╝ ╚═╝╚══════╝

"""


def get_pytoil_version() -> str:
    """
    Gets the current pytoil __version__.

    Returns:
        str: v__version__
    """
    return f"v{__version__}"


def get_python_version() -> str:
    """
    Returns the python version used to call pytoil.

    Returns:
        str: Python version.
    """
    return sys.version.replace("\n", "").strip()


def version_callback(value: bool) -> None:
    """
    Callback responsible for printing the version info.
    """

    version_dict: Dict[str, str] = {
        "python version": get_python_version(),
        "pytoil version": get_pytoil_version(),
        "platform": platform.system(),
        "OS": platform.platform(),
        "arch": platform.machine(),
    }

    if value:
        typer.echo(LOGO)

        for key, val in version_dict.items():

            version_start = typer.style(f"{key}: ", fg=typer.colors.CYAN)
            version_msg = version_start + f"{val}"

            typer.echo(version_msg)
        raise typer.Exit(0)



from enum import Enum

from pytoil.starters.go import GoStarter
from pytoil.starters.python import PythonStarter
from pytoil.starters.rust import RustStarter

__all__ = [
    "PythonStarter",
    "GoStarter",
    "RustStarter",
]


class Starter(str, Enum):
    """
    Choice of starter templates.
    """

    python = "python"
    go = "go"
    rust = "rust"
    none = "none"



"""
The `pytoil show` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""


import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config

app = typer.Typer(name="show")


# Callback for documentation only
@app.callback()
def show() -> None:
    """
    View your local/remote projects.

    The show command provides an easy way of listing of the projects you have locally
    in your configured development directory and/or of those you have on GitHub
    (known in pytoil-land as 'remote' projects).

    Local projects will be the names of subdirectories in your configured projects
    directory.

    The remote projects listed here will be those owned by you on GitHub.

    The "--count/-c" flag can be used if you just want to see a count.
    """


@app.command()
def local(
    count: bool = typer.Option(
        False,
        "--count",
        "-c",
        help="Display a count of local projects.",
        show_default=False,
    )
) -> None:
    """
    Show your local projects.

    Show the projects you have locally in your configured
    projects directory.

    If the "--count/-c" flag is used, you will simply be
    shown the count of local projects.

    Examples:

    $ pytoil show local

    $ pytoil show local --count
    """
    config = Config.from_file()
    local_projects = utils.get_local_projects(path=config.projects_dir)

    if not local_projects:
        msg.warn("You don't have any local projects yet!", exits=0)

    typer.secho("\nLocal Projects:\n", fg=typer.colors.CYAN, bold=True)

    if count:
        typer.echo(f"You have {len(local_projects)} local projects")
    else:
        for project in sorted(local_projects, key=str.casefold):
            typer.echo(f"- {project}")


@app.command()
def remote(
    count: bool = typer.Option(
        False,
        "--count",
        "-c",
        help="Display a count of remote projects.",
        show_default=False,
    )
) -> None:
    """
    Show your remote projects.

    Show the projects that you own on GitHub.

    These mayinclude some you already have locally.
    Use 'show diff' to see the difference between local and remote.

    The "--count/-c" flag can be used to show a count of remote projects.

    Examples:

    $ pytoil show remote

    $ pytoil show remote --count
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    try:
        with msg.loading("Fetching your remote projects..."):
            remote_projects = api.get_repo_names()
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        if not remote_projects:
            msg.warn("You dont have any projects on GitHub yet!", exits=0)

        typer.secho("\nRemote Projects:\n", fg=typer.colors.CYAN, bold=True)

        if count:
            typer.echo(f"You have {len(remote_projects)} remote projects.")
        else:
            for project in sorted(remote_projects, key=str.casefold):
                typer.echo(f"- {project}")


# Can't call it 'all', python keyword
@app.command(name="all")
def all_(
    count: bool = typer.Option(
        False,
        "--count",
        "-c",
        help="Display a count of all projects.",
        show_default=False,
    )
) -> None:
    """
    Show all your projects, grouped by local and remote.

    The "--count/-c" flag will show a count of local/remote projects.

    Examples:

    $ pytoil show all

    $ pytoil show all --count
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    try:
        with msg.loading("Fetching your projects..."):
            local_projects = utils.get_local_projects(path=config.projects_dir)
            remote_projects = api.get_repo_names()
            n_locals = len(local_projects)
            n_remotes = len(remote_projects)
            total = n_locals + n_remotes
            local_pct = n_locals / total * 100
            remote_pct = n_remotes / total * 100

    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        typer.secho("\nLocal Projects:\n", fg=typer.colors.CYAN, bold=True)

        if not local_projects:
            msg.warn("You don't have any local projects yet!", exits=0)
        elif count:
            typer.echo(
                f"You have {n_locals} local projects. {local_pct:.2f}% of total."
            )
        else:
            for project in sorted(local_projects, key=str.casefold):
                typer.echo(f"- {project}")

        typer.secho("\nRemote Projects:\n", fg=typer.colors.CYAN, bold=True)

        if not remote_projects:
            msg.warn("You don't have any projects on GitHub yet!", exits=0)
        elif count:
            typer.echo(
                f"You have {n_remotes} remote projects. {remote_pct:.2f}% of total."
            )
        else:
            for project in sorted(remote_projects, key=str.casefold):
                typer.echo(f"- {project}")


@app.command()
def diff(
    count: bool = typer.Option(
        False,
        "--count",
        "-c",
        help="Display a count of the diff.",
        show_default=False,
    )
) -> None:
    """
    Show the difference in local/remote projects.

    Show the projects that you own on GitHub but do not
    have locally.

    The "--count/-c" flag will show a count of the difference.

    Examples:

    $ pytoil show diff

    $ pytoil show diff --count
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    try:
        with msg.loading("Calculating difference..."):
            local_projects = utils.get_local_projects(path=config.projects_dir)
            remote_projects = api.get_repo_names()

            diff = remote_projects.difference(local_projects)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        if not diff:
            msg.info("Your local and remote projects are in sync!")
            msg.good("Nothing to show!", exits=0)

        typer.secho("\nDiff: Remote - Local\n", fg=typer.colors.CYAN, bold=True)

        if count:
            typer.echo(
                f"You have {len(diff)} projects on GitHub that aren't available locally."  # noqa: E501
            )
        else:
            for project in sorted(diff, key=str.casefold):
                typer.echo(f"- {project}")



"""
The pytoil root CLI command.

This module simply instantiates the base pytoil CLI command
and attaches it's direct subcommands and it's subcommand
groups.

Author: Tom Fleet
Created: 18/06/2021
"""


import typer

from pytoil.cli import checkout, config, docs, gh, info, new, pull, remove, show
from pytoil.cli.version import version_callback

# Create the root app
app = typer.Typer(name="pytoil", no_args_is_help=True)

# Attach immediate child commands
# this is a bit hacky but it lets us have one command per file
# which is a nicer design
app.registered_commands += (
    checkout.app.registered_commands
    + new.app.registered_commands
    + docs.app.registered_commands
    + gh.app.registered_commands
    + info.app.registered_commands
    + remove.app.registered_commands
    + pull.app.registered_commands
)

# Add the sub command groups
# these are commands with their own subcommands
app.add_typer(show.app, name="show")
app.add_typer(config.app, name="config")


# Callback for documentation only
@app.callback()
def main(
    version: bool = typer.Option(
        None,
        "--version",
        "-V",
        help="Display pytoil version.",
        callback=version_callback,
        is_eager=True,
        show_default=False,
    )
) -> None:
    """
    Helpful CLI to automate the development workflow.

    - Create and manage your local and remote projects

    - Build projects from cookiecutter templates.

    - Easily create/manage virtual environments.

    - Minimal configuration required.
    """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytoil',
 'pytoil.api',
 'pytoil.cli',
 'pytoil.config',
 'pytoil.environments',
 'pytoil.git',
 'pytoil.repo',
 'pytoil.starters',
 'pytoil.vscode']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'cookiecutter>=1.7.3,<2.0.0',
 'httpx>=0.18.2,<0.19.0',
 'pydantic>=1.8.2,<2.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer[all]>=0.3.2,<0.4.0',
 'wasabi>=0.8.2,<0.9.0']

entry_points = \
{'console_scripts': ['pytoil = pytoil.cli.root:app']}

setup_kwargs = {
    'name': 'pytoil',
    'version': '0.8.1',
    'description': 'CLI to automate the development workflow.',
    'long_description': "![logo](./docs/img/logo.png)\n\n[![License](https://img.shields.io/github/license/FollowTheProcess/pytoil)](https://github.com/FollowTheProcess/pytoil)\n[![PyPI](https://img.shields.io/pypi/v/pytoil.svg?logo=python)](https://pypi.python.org/pypi/pytoil)\n[![GitHub](https://img.shields.io/github/v/release/FollowTheProcess/pytoil?logo=github&sort=semver)](https://github.com/FollowTheProcess/pytoil)\n[![Code Style](https://img.shields.io/badge/code%20style-black-black)](https://github.com/FollowTheProcess/pytoil)\n[![CI](https://github.com/FollowTheProcess/pytoil/workflows/CI/badge.svg)](https://github.com/FollowTheProcess/pytoil/actions?query=workflow%3ACI)\n[![Coverage](docs/img/coverage.svg)](https://github.com/FollowTheProcess/pytoil)\n\n*pytoil is a small, helpful CLI to help developers manage their local and remote projects with ease!*\n\n* **Source Code**: [https://github.com/FollowTheProcess/pytoil](https://github.com/FollowTheProcess/pytoil)\n\n* **Documentation**: [https://FollowTheProcess.github.io/pytoil/](https://FollowTheProcess.github.io/pytoil/)\n\n:warning: pytoil is still in Alpha and as such, the API may change without deprecation notices.\n\n## What is it?\n\n`pytoil` is a handy tool that helps you stay on top of all your projects, remote or local. It's primarily aimed at python developers but you could easily use it to manage any project!\n\npytoil is:\n\n* Easy to use :white_check_mark:\n* Easy to configure :white_check_mark:\n* Safe (it won't edit your repos at all) :white_check_mark:\n* Useful! (I hope :smiley:)\n\nSay goodbye to janky bash scripts :wave:\n\n## Installation\n\nAs pytoil is a CLI, I recommend [pipx]\n\n```shell\npipx install pytoil\n```\n\nOr just pip (but must be globally available)\n\n```shell\npip install pytoil\n```\n\n## Quickstart\n\n`pytoil` is super easy to get started with.\n\nAfter installation just run\n\n```shell\n$ pytoil config\n\nNo config file yet!\nMaking you a default one...\n```\n\nThis will create a default config file which can be found at `~/.pytoil.yml`. See the [docs] for what information you need to put in here.\n\nDon't worry though, there's only a few options to configure! :sleeping:\n\nAfter that you're good to go! You can do things like:\n\n#### See your local and remote projects\n\n```shell\npytoil show all\n```\n\n#### See which ones you have on GitHub, but not on your computer\n\n```shell\npytoil show diff\n```\n\n#### Easily grab a project, regardless of where it is\n\n```shell\npytoil checkout my_project\n```\n\n#### Create a new project and virtual environment in one go\n\n```shell\npytoil new my_project --venv venv\n\n```\n\n#### And even do this from a [cookiecutter] template\n\n```shell\npytoil new my_project --venv venv --cookie https://github.com/some/cookie.git\n```\n\nCheck out the [docs] for more :boom:\n\n### Credits\n\nThis package was created with [cookiecutter] and the [FollowTheProcess/poetry_pypackage] project template.\n\n`pytoil` has been built on top of these fantastic projects:\n\n* [Typer]\n* [cookiecutter]\n* [wasabi]\n* [httpx]\n* [pydantic]\n\n[pipx]: https://pipxproject.github.io/pipx/\n[cookiecutter]: https://cookiecutter.readthedocs.io/en/1.7.2/\n[docs]: https://FollowTheProcess.github.io/pytoil/\n[FollowTheProcess/poetry_pypackage]: https://github.com/FollowTheProcess/poetry_pypackage\n[Typer]: https://typer.tiangolo.com\n[wasabi]: https://github.com/ines/wasabi\n[httpx]: https://www.python-httpx.org\n[pydantic]: https://pydantic-docs.helpmanual.io\n",
    'author': 'Tom Fleet',
    'author_email': 'tomfleet2018@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FollowTheProcess/pytoil',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
The `pytoil pull` subcommand group.

Author: Tom Fleet
Created: 18/06/2021
"""

from typing import List, Set

import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.git import Git
from pytoil.repo import Repo

app = typer.Typer()


@app.command()
def pull(
    projects: List[str] = typer.Argument(
        None,
        help="Name of the project(s) to pull down.",
        show_default=False,
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force pull without confirmation.",
        show_default=False,
    ),
    all_: bool = typer.Option(
        False,
        "--all",
        "-a",
        help="Pull down all of your projects.",
        show_default=False,
    ),
) -> None:
    """
    Pull down your remote projects.

    The pull command provides easy methods for pulling down remote projects.

    It is effectively a nice wrapper around git clone but you don't have to
    worry about urls or what your cwd is, pull will grab your remote projects
    by name and clone them to your configured projects directory.

    You can also use pull to batch clone multiple repos, even all of them ("--all/-a")
    if you're into that sorta thing.

    Any remote project that already exists locally will be skipped and none of
    your local projects are changed in any way. pytoil will only pull down
    those projects that don't already exist locally.

    It's very possible to accidentally clone a lot of repos when using pull so
    you will be prompted for confirmation before pytoil does anything.

    The "--force/-f" flag can be used to override this confirmation prompt if
    desired.

    Examples:

    $ pytoil pull project1 project2 project3

    $ pytoil pull project1 project2 project3 --force

    $ pytoil pull --all

    $ pytoil pull --all --force
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    api = API(username=config.username, token=config.token)

    if not projects and not all_:
        msg.warn(
            "If not using the '--all' flag, you must specify projects to pull.",
            exits=1,
            spaced=True,
        )

    try:
        not_found = False
        with msg.loading("Calculating difference..."):
            local_projects = utils.get_local_projects(path=config.projects_dir)
            remote_projects = api.get_repo_names()

            specified_remotes = remote_projects if all_ else set(projects)

            for project in projects:
                if project not in remote_projects:
                    not_found = True
                    not_found_project = project
                    break
                break

        if not_found:
            # If we don't break out of the with block
            # the spinner will go forever
            msg.warn(
                f"{not_found_project!r} not found on GitHub. Was it a typo?",
                spaced=True,
                exits=1,
            )

        diff = specified_remotes.difference(local_projects)
    except httpx.HTTPStatusError as err:
        utils.handle_http_status_errors(error=err)
    else:

        pull_diff(
            diff=diff, remotes=remote_projects, force=force, config=config, git=Git()
        )


def pull_diff(
    diff: Set[str], remotes: Set[str], force: bool, config: Config, git: Git
) -> None:
    """
    Helper function to help the pull CLI command with repeated logic.

    Args:
        diff (Set[str]): The difference to clone.
        remotes (Set[str]): Set of remote projects.
        force (bool): The value for the '--force' flag.
        config (Config): The Config object.
    """

    if not diff:
        msg.info("Your local and remote projects are in sync.")
        msg.good("Nothing to pull!", exits=0)

    if not remotes:
        msg.warn("You don't have any remote projects to pull!", spaced=True, exits=1)

    if not force:
        if len(diff) <= 5:
            # A managable length to display each one
            typer.confirm(
                f"This will pull down {', '.join(diff)}. Are you sure?", abort=True
            )
        else:
            # Too many to show to look nice, just show number
            typer.confirm(
                f"This will pull down {len(diff)} projects. Are you sure?", abort=True
            )

    # If we get here, we're good to go

    for project in diff:
        # Make a Repo object for each
        repo = Repo(
            owner=config.username,
            name=project,
            local_path=config.projects_dir.joinpath(project),
        )

        msg.info(f"Cloning {project!r}", spaced=True)
        git.clone(url=repo.clone_url, cwd=config.projects_dir)

    msg.good("Done!", spaced=True)



"""
The pytoil checkout command.

Author: Tom Fleet
Created: 25/06/2021
"""

import re
import time
from typing import Optional

import httpx
import typer
from wasabi import msg

from pytoil.api import API
from pytoil.cli import utils
from pytoil.config import Config
from pytoil.environments.abstract import Environment
from pytoil.exceptions import (
    EnvironmentAlreadyExistsError,
    ExternalToolNotInstalledException,
)
from pytoil.git import Git
from pytoil.repo import Repo
from pytoil.vscode import VSCode

app = typer.Typer()

# The username/repo pattern
USER_REPO_REGEX = re.compile(r"^([A-Za-z0-9_.-])+/([A-Za-z0-9_.-])+$")

# The bare 'project' pattern
PROJECT_REGEX = re.compile(r"^([A-Za-z0-9_.-])+$")


@app.command()
def checkout(
    project: str = typer.Argument(
        ...,
        help="Name of the project to checkout.",
    ),
    venv: bool = typer.Option(
        False,
        "--venv",
        "-v",
        help="Attempt to auto-create a virtual environment.",
        show_default=False,
    ),
) -> None:
    """
    Checkout an existing development project.

    The checkout command lets you easily resume work on an existing project,
    whether that project is available locally in your configured projects
    directory, or if it is on GitHub.

    If pytoil finds your project locally, and you have enabled VSCode in your
    config file it will open it for you. If not, it will just tell you it already
    exists locally and where to find it.

    If your project is on your GitHub, pytoil will clone it for you and then open it
    (or tell you where it cloned it if you dont have VSCode set up).

    Finally, if checkout can't find a match after searching locally and on GitHub,
    it will prompt you to use 'pytoil new' to create a new one.

    If you pass the shorthand to someone elses repo e.g. 'someoneelse/repo' pytoil
    will detect this and automatically create a fork of this repo for you. Forking
    happens asynchronously so we can't clone your fork straight away. Give it a
    few seconds then a 'pytoil checkout repo' will bring it down as normal.

    You can also ask pytoil to automatically create a virtual environment on
    checkout with the '--venv/-v' flag. This only happens for projects pulled down
    from GitHub to avoid accidentally screwing up local projects.

    If the '--venv/-v' flag is used, pytoil will look at your project to try and detect
    which type of environment to create e.g. conda, flit, poetry, standard python etc.

    The '--venv/-v' flag will also attempt to detect if the project you're checking out
    is a python package, in which case it will install it's requirements into the
    created environment.

    More info about this can be found in the documentation.

    Examples:

    $ pytoil checkout my_project

    $ pytoil checkout my_project --venv

    $ pytoil checkout someoneelse/project
    """
    config = Config.from_file()
    utils.warn_if_no_api_creds(config)

    # Setup the objects required
    api = API(username=config.username, token=config.token)
    repo = Repo(
        owner=config.username,
        name=project,
        local_path=config.projects_dir.joinpath(project),
    )
    code = VSCode(root=repo.local_path)
    git = Git()

    if bool(USER_REPO_REGEX.match(project)):
        # We've matched the "user/repo" pattern, meaning the user
        # wants to create a fork
        owner, name = project.split("/")
        fork_repo(
            owner=owner,
            name=name,
            api=api,
            config=config,
            git=git,
            venv=venv,
        )
        msg.good("Done!", spaced=True, exits=0)

    elif bool(PROJECT_REGEX.match(project)):
        # User has just passed a single project

        if repo.exists_local():
            # No environment or git stuff here, chances are if it exists locally
            # user has already done all this stuff
            checkout_local(repo=repo, code=code, config=config)

        elif repo.exists_remote(api=api):
            msg.info(f"{repo.name!r} found on GitHub. Cloning...", spaced=True)
            git.clone(url=repo.clone_url, check=True, cwd=config.projects_dir)

            env = repo.dispatch_env()

            if venv:
                handle_venv_creation(env=env, config=config, code=code)

            if config.vscode:
                msg.info(f"Opening {repo.name!r} in VSCode.", spaced=True)
                code.open()

        else:
            msg.warn(
                title=f"{repo.name!r} not found locally or on GitHub!",
                text=f"Does it exist? If not, create a new project with 'pytoil new {repo.name}'.",  # noqa: E501
                exits=1,
            )

    else:
        # User has passed garbage
        msg.fail(
            f"Argument: {project} did not match any valid syntax.",
            text="Valid syntax is 'project' for a local or remote project you own, "
            "or 'user/project' to fork a repo you don't own.",
            exits=1,
            spaced=True,
        )


def fork_repo(
    owner: str, name: str, api: API, config: Config, git: Git, venv: bool
) -> None:
    """
    Forks the passed repo, clones it, sets the upstream
    and informs the user along the way.

    Args:
        owner (str): Owner of the repo to be forked.
        name (str): Name of the repo to be forked.
        api (API): Configured API object.
        config (Config): Configured Config object.
        git (Git): The Git object.
        venv (bool): Whether or not the user wants a virtual environment.
    """
    # TODO: Lots of duplicated code here with the rest of checkout
    # simplify this so it integrates a bit better

    if owner == config.username:
        msg.warn(
            "You don't need the '/' when checking out a repo you own!",
            spaced=True,
            exits=1,
        )

    msg.info(f"'{owner}/{name}' belongs to {owner!r}.", spaced=True)
    with msg.loading(f"Forking '{owner}/{name}'..."):
        try:
            api.create_fork(owner=owner, repo=name)
        except httpx.HTTPStatusError as err:
            utils.handle_http_status_errors(err)

        # Forking happens asynchronously
        # see https://docs.github.com/en/rest/reference/repos#create-a-fork
        # So here we naively just wait a few seconds before trying to clone the fork
        # this may need to change and there is most likely a better way of doing this
        time.sleep(3)

    # Now we check if the user's fork exists
    # If not, we report that to the user and exit
    fork = Repo(
        owner=config.username, name=name, local_path=config.projects_dir.joinpath(name)
    )
    code = VSCode(root=fork.local_path)
    if not fork.exists_remote(api=api):
        msg.warn(
            "Fork not available yet",
            text="Forking happens asynchronously so this is normal. "
            "Give it a few more seconds and try checking it out again.",
            exits=1,
        )

    # If we get here, the fork was a success
    msg.info(f"Cloning your fork: {config.username}/{name}.", spaced=True)
    git.clone(url=fork.clone_url, cwd=config.projects_dir)

    # Set upstream
    msg.info("Setting 'upstream' to original repo.", spaced=True)
    git.set_upstream(owner=owner, repo=name, path=fork.local_path)

    env = fork.dispatch_env()

    if venv:
        handle_venv_creation(env=env, config=config, code=code)

    if config.vscode:
        msg.info(f"Opening {fork.name!r} in VSCode.")
        code.open()


def checkout_local(repo: Repo, code: VSCode, config: Config) -> None:
    """
    Helper function to checkout a local repo.

    Args:
        repo (Repo): Repo to checkout.
        code (VSCode): VSCode instance to handle opening.
        config (Config): Configured Config object.
    """

    msg.info(f"{repo.name!r} available locally.", spaced=True)

    if config.vscode:
        msg.text(f"Opening {repo.name!r} in VSCode.")
        code.open()


def handle_venv_creation(
    env: Optional[Environment], config: Config, code: VSCode
) -> None:
    """
    Handles the automatic detection and/or creation of
    python virtual environments based on detected repo
    contents and reports to the user.

    Args:
        env (Optional[Environment]): The Environment object
        config (Config): Populated Config object.
        code (VSCode): VSCode object.
    """
    if not env:
        msg.warn(
            "Unable to auto-detect required environent. Skipping.",
            spaced=True,
        )
    else:
        msg.info(
            f"Auto creating virtual environment using: {env.info_name!r}",
            spaced=True,
        )
        if env.info_name == "conda":
            msg.text("Note: Conda environments can take up to a few minutes to create.")

        try:
            with msg.loading("Working..."):
                env.install_self()
        except ExternalToolNotInstalledException:
            msg.fail(title=f"{env.info_name!r} not installed", spaced=True, exits=1)
        except EnvironmentAlreadyExistsError:
            msg.warn(
                title="Environment already exists!",
                text="No need to create a new one. Skipping.",
            )
        finally:
            # Because the first exception will exit, this is okay
            # to call as finally
            if config.vscode:
                code.set_workspace_python(python_path=env.executable)
